<?php

namespace Drupal\cl_components\Plugin\Field\FieldType;

use Drupal\cl_components\Component\Component;
use Drupal\cl_components\Component\ComponentDiscovery;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemBase;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\TypedData\DataDefinition;

/**
 * Defines the 'cl_components_style_selector' field type.
 *
 * @FieldType(
 *   id = "cl_components_style_selector",
 *   label = @Translation("Style Selector"),
 *   category = @Translation("General"),
 *   default_widget = "cl_components_style_selector",
 *   default_formatter = "cl_components_style_selector_table"
 * )
 */
class StyleSelectorItem extends FieldItemBase {

  const MODULE = 'module';

  const VARIANT = 'variant';

  /**
   * {@inheritdoc}
   */
  public static function defaultFieldSettings() {
    $settings = [
      'features' => [
        static::VARIANT => static::VARIANT,
      ],
      'forbidden' => [],
      'allowed' => [],
    ];
    return $settings + parent::defaultFieldSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function fieldSettingsForm(array $form, FormStateInterface $form_state) {
    $settings = $this->getSettings();

    $element['features'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Supported features'),
      '#description' => $this->t('The module is mandatory, but the variant support is optional for each field instance.'),
      '#default_value' => $settings['features'],
      '#options' => [
        static::VARIANT => $this->t('Variant'),
      ],
    ];
    $options = static::allAllowedModuleValues();
    $element['forbidden'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Forbidden modules'),
      '#description' => $this->t('When showing the module dropdown, all modules will be displayed except the modules selected here. This is used to limit the modules available for a given content type.'),
      '#options' => $options,
      '#default_value' => $settings['forbidden'],
      '#attributes' => ['class' => ['long-checkboxes-list']],
    ];
    $element['allowed'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Allowed modules'),
      '#description' => $this->t('When showing the module dropdown, only the modules selected here will be displayed as an option. <strong>NOTE:</strong> "forbidden" modules will not be shown even when allowed here.'),
      '#options' => $options,
      '#default_value' => $settings['allowed'],
      '#attributes' => ['class' => ['long-checkboxes-list']],
    ];
    $element['#attached']['library'][] = 'cl_components/cl_components_style_selector';

    return $element;
  }

  /**
   * Returns all the allowed values for 'module' sub-field.
   *
   * @param array $allowed
   *   The allowed modules.
   * @param array $forbidden
   *   The forbidden modules.
   *
   * @return array
   *   The list of allowed values.
   */
  public static function allAllowedModuleValues(array $allowed = [], array $forbidden = []): array {
    $options = static::allAllowedModuleValuesSelect($allowed, $forbidden);
    return array_reduce(
      $options,
      static function (array $options, array $group) {
        return array_merge($options, $group);
      },
      []
    );
  }

  /**
   * Returns all the allowed values for 'module' sub-field grouped by letter.
   *
   * @param array $allowed
   *   The allowed modules.
   * @param array $forbidden
   *   The forbidden modules.
   *
   * @return array
   *   The list of allowed values.
   */
  public static function allAllowedModuleValuesSelect(array $allowed = [], array $forbidden = []): array {
    $modules = \Drupal::service(ComponentDiscovery::class)->findAllModules();
    $options = [];
    foreach ($modules as $component) {
      assert($component instanceof Component);
      $metadata = $component->getMetadata();
      $machine_name = $metadata->getMachineName();
      // If there are allowed modules and this is not one of them, continue.
      if (!empty($allowed) && !in_array($machine_name, $allowed, TRUE)) {
        continue;
      }
      // If this is a forbidden module, continue.
      if (in_array($machine_name, $forbidden, TRUE)) {
        continue;
      }
      if (!isset($options[$metadata->getGroup()])) {
        $options[$metadata->getGroup()] = [];
      }
      $name = $component->getMetadata()->getName();
      $options[$metadata->getGroup()][$machine_name] = $name;
    }
    ksort($options);
    foreach ($options as $group => $opts) {
      natsort($options[$group]);
    }
    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function isEmpty() {
    return $this->module === NULL
      && $this->variant === NULL;
  }

  /**
   * {@inheritdoc}
   */
  public static function propertyDefinitions(FieldStorageDefinitionInterface $field_definition) {
    $properties['module'] = DataDefinition::create('string')
      ->setLabel(t('Module'));
    $properties['variant'] = DataDefinition::create('string')
      ->setLabel(t('Variant'));

    return $properties;
  }

  /**
   * {@inheritdoc}
   */
  public function getConstraints() {
    $constraints = parent::getConstraints();

    $options['module']['AllowedValues'] = array_keys(static::allAllowedModuleValues());
    $options['module']['NotBlank'] = [];

    $constraint_manager = \Drupal::typedDataManager()
      ->getValidationConstraintManager();
    $constraints[] = $constraint_manager->create('ComplexData', $options);
    return $constraints;
  }

  /**
   * {@inheritdoc}
   */
  public static function schema(FieldStorageDefinitionInterface $field_definition) {
    return [
      'columns' => [
        'module' => [
          'type' => 'varchar',
          'length' => 255,
        ],
        'variant' => [
          'type' => 'varchar',
          'length' => 255,
        ],
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public static function generateSampleValue(FieldDefinitionInterface $field_definition) {
    return [
      'module' => array_rand(static::allAllowedModuleValues()),
      'variant' => 'my-variant',
    ];
  }

  /**
   * Returns allowed values for 'module' sub-field.
   *
   * @return array
   *   The list of allowed values.
   */
  public function allowedModuleValues(): array {
    return array_reduce(
      $this->allowedModuleValuesSelect(),
      static function (array $options, array $group) {
        return array_merge($options, $group);
      },
      []
    );
  }

  /**
   * Returns allowed values for 'module' sub-field when using a select element.
   *
   * @return array
   *   The list of allowed values.
   */
  public function allowedModuleValuesSelect(): array {
    $settings = $this->getSettings();
    $allowed = array_keys(array_filter($settings['allowed'] ?? []));
    $forbidden = array_keys(array_filter($settings['forbidden'] ?? []));
    return static::allAllowedModuleValuesSelect($allowed, $forbidden);
  }

}
