<?php

namespace Drupal\cl_components\Plugin\Field\FieldWidget;

use Drupal\cl_components\Component\ComponentDiscovery;
use Drupal\cl_components\Exception\InvalidComponentException;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\WidgetBase;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Validator\ConstraintViolationInterface;

/**
 * Defines the 'cl_components_style_selector' field widget.
 *
 * @FieldWidget(
 *   id = "cl_components_style_selector",
 *   label = @Translation("Style Selector"),
 *   field_types = {"cl_components_style_selector"},
 * )
 */
class StyleSelectorWidget extends WidgetBase {

  /**
   * The discovery.
   *
   * @var \Drupal\cl_components\Component\ComponentDiscovery
   */
  private $componentDiscovery;

  /**
   * The path wher Drupal is installed.
   *
   * @var string
   */
  private $appRoot;

  /**
   * {@inheritdoc}
   */
  public function __construct($plugin_id, $plugin_definition, FieldDefinitionInterface $field_definition, array $settings, array $third_party_settings, ComponentDiscovery $component_discovery, string $app_root) {
    parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $third_party_settings);
    $this->componentDiscovery = $component_discovery;
    $this->appRoot = $app_root;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $component_discovery = $container->get(ComponentDiscovery::class);
    assert($component_discovery instanceof ComponentDiscovery);
    return new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'], $configuration['third_party_settings'],
      $component_discovery,
      $container->getParameter('app.root')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function formElement(FieldItemListInterface $items, $delta, array $element, array &$form, FormStateInterface $form_state) {
    $item = $items[$delta] ?? NULL;
    $all_parents = array_merge(
      $element['#field_parents'],
      [$items->getName(), $delta, 'module']
    );
    $user_input = $form_state->getUserInput();
    $selected_module = NestedArray::getValue($user_input, $all_parents);
    $selected_module = !is_null($selected_module)
      ? $selected_module
      : ($item->module ?? NULL);
    $element['module'] = [
      '#type' => 'select',
      '#title' => $this->t('Module'),
      '#options' => ['' => $this->t('- Select a value -')] + $item->allowedModuleValuesSelect(),
      '#default_value' => $selected_module,
      '#ajax' => [
        'wrapper' => 'style-selector-wrapper',
        'method' => 'replaceWith',
        'callback' => [$this, 'ajaxDataHandler'],
      ],
    ];

    if ($selected_module) {
      $features = array_keys(array_filter($items->getSetting('features')));
      if (in_array('variant', $features, TRUE)) {
        $current_variant = $items[$delta]->variant;
        $allowed_variants = $item->allowedVariantValues($selected_module);
        $element['variant'] = [
          '#type' => 'select',
          '#title' => $this->t('Variant'),
          '#options' => $allowed_variants,
          '#default_value' => isset($allowed_variants[$current_variant]) ? $current_variant : 'variantNone',
        ];
      }
      $element['documentation'] = $this->buildComponentInfo($selected_module);
    }

    $element['#theme_wrappers'] = ['container', 'form_element'];
    $element['#attributes']['class'][] = 'container-inline';
    $element['#attributes']['id'][] = 'container-inline-style-selector';
    $element['#prefix'] = '<div id="style-selector-wrapper">';
    $element['#suffix'] = '</div>';
    $element['#attributes']['class'][] = 'hc-fractal-style-selector-elements';
    $element['#attached']['library'][] = 'cl_components/cl_components_style_selector';

    return $element;
  }

  /**
   * Builds the component information.
   *
   * @param string $machine_name
   *   The machine name of the component.
   *
   * @throws \Drupal\cl_components\Exception\ComponentNotFoundException
   *   When the component cannot be found.
   */
  private function buildComponentInfo(string $machine_name): array {
    try {
      $component = $this->componentDiscovery->find($machine_name);
    }
    catch (InvalidComponentException $exception) {
      return [];
    }
    $metadata = $component->getMetadata();
    $filename = $metadata->getThumbnailPath();
    $relative = str_replace($this->appRoot, '', $filename);
    return [
      '#type' => 'details',
      '#attributes' => ['class' => ['module-info']],
      '#title' => $this->t(
        'About %module <span class="component-status @status"></span>',
        ['%module' => $metadata->getName(), '@status' => $metadata->getStatus()]
      ),
      'thumbnail' => file_exists($filename) ? [
        '#theme' => 'image',
        '#uri' => drupal_get_filename('module', 'cl_components', $relative),
        '#alt' => $this->t('Thumbnail for @module', ['@module' => $metadata->getName()]),
      ] : [],
      'docs' => [
        '#type' => 'container',
        'documentation' => ['#markup' => $metadata->getDocumentation()],
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function errorElement(array $element, ConstraintViolationInterface $violation, array $form, FormStateInterface $form_state) {
    return isset($violation->arrayPropertyPath[0]) ? $element[$violation->arrayPropertyPath[0]] : $element;
  }

  /**
   * {@inheritdoc}
   */
  public function massageFormValues(array $values, array $form, FormStateInterface $form_state) {
    foreach ($values as $delta => $value) {
      $module = $value['module'] ?? '';
      if ($module === '') {
        $values[$delta]['module'] = NULL;
        $values[$delta]['variant'] = NULL;
      }
      $variant = $value['variant'] ?? '';
      if ($variant === '') {
        $values[$delta]['variant'] = NULL;
      }
    }
    return $values;
  }

  /**
   * Handles the AJAX data.
   *
   * @param array $form
   *   The form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form interface.
   *
   * @return array|mixed|null
   *   The results.
   */
  public function ajaxDataHandler(array &$form, FormStateInterface $form_state) {
    // Pop the last element so we render the whole triad.
    $array_parents = array_slice(
      $form_state->getTriggeringElement()['#array_parents'],
      0,
      -1
    );
    return NestedArray::getValue($form, $array_parents);
  }

}
