<?php

namespace Drupal\cl_components\Twig;

use Drupal\cl_components\Component\ComponentDiscovery;
use Drupal\cl_components\Exception\TemplateNotFoundException;
use Drupal\cl_components\Service\ComponentRenderer;
use Drupal\cl_components\Twig\tag\ComponentTokenParser;
use Twig\Compiler;
use Twig\Extension\AbstractExtension;
use Twig\TwigFunction;

/**
 * The twig extension so Drupal can recognize the new code.
 */
class TwigExtension extends AbstractExtension {

  /**
   * The renderer.
   *
   * @var \Drupal\cl_components\Service\ComponentRenderer
   */
  private ComponentRenderer $renderer;

  /**
   * The discovery.
   *
   * @var \Drupal\cl_components\Component\ComponentDiscovery
   */
  private ComponentDiscovery $discovery;

  /**
   * TwigComponentExtension constructor.
   *
   * @param \Drupal\cl_components\Service\ComponentRenderer $component_renderer
   *   The component renderer.
   */
  public function __construct(
    ComponentRenderer $component_renderer,
    ComponentDiscovery $discovery
  ) {
    $this->renderer = $component_renderer;
    $this->discovery = $discovery;
  }

  /**
   * {@inheritdoc}
   */
  public function getFunctions() {
    return [
      new TwigFunction('cl_components_additional_context', [$this, 'addAdditionalContext'], ['needs_context' => TRUE]),
    ];
  }

  /**
   * Appends additional context to the template based on the template id.
   *
   * @param array &$context
   *   The context.
   * @param string $component_id
   *   The component ID.
   * @param string $variant
   *   The variant.
   *
   * @throws \Drupal\cl_components\Exception\TemplateNotFoundException|\Drupal\cl_components\Exception\ComponentNotFoundException
   */
  public function addAdditionalContext(array &$context, string $component_id, string $variant = '') {
    $component = $this->discovery->find($component_id);
    if (!empty($variant) && !in_array($variant, $component->getVariants())) {
      $message = sprintf(
        'Unable to render variant "%s". This variant is not declared in the metadata.json for this component.',
        $variant
      );
      throw new TemplateNotFoundException($message);
    }
    $context = array_merge(
      $context,
      $component->additionalRenderContext($variant)
    );
  }
  /**
   * {@inheritdoc}
   */
  public function getTokenParsers(): array {
    return [new ComponentTokenParser('inject')];
  }

  /**
   * {@inheritdoc}
   */
  public function getNodeVisitors(): array {
    return [new ComponentNodeVisitor()];
  }

  /**
   * Gets the renderer.
   *
   * @return \Drupal\cl_components\Service\ComponentRenderer
   *   The renderer.
   */
  public function getRenderer(): ComponentRenderer {
    return $this->renderer;
  }

  /**
   * Creates a Twig compiler.
   *
   * @return \Twig\Compiler
   *   The compiler.
   */
  public function createCompiler(): Compiler {
    return new Compiler($this->renderer->getTwigEnvironment());
  }

}
